package org.garret.heap;

import java.util.*;
import java.lang.reflect.*;

public class HeapInfo { 
    /**
     * Add object to the list of root objects
     * @param obj root object
     */
    public void addRoot(Object obj) {
        rootObjects.add(obj);
    }
    
    /**
     * Add all static components of the class to the list of root objects
     * @param cls class which static components will be added to the list of root objects
     */
    public void addRootClass(Class cls) { 
        rootClasses.add(cls);
    }

    /**
     * Build object closure - collect heap info
     */
    public void refresh() {
        objects.clear();
        greyList = null;
        blackList = null;
        for (int i = 0, n = rootObjects.size(); i < n; i++) { 
            add(rootObjects.get(i), this);
        }
        for (int i = 0, n = rootClasses.size(); i < n; i++) { 
            addStaticFields((Class)rootClasses.get(i));
        }
        mark();
    }
    
    /**
     * Get list of objects referencing this object
     * @param obj referened object
     * @return array of objects referencing this object
     */
    public Object[] referencedBy(Object obj) { 
        ObjectHeader hdr = (ObjectHeader)objects.get(obj);
        if (hdr != null) { 
            return hdr.referencedBy.toArray();
        }
        return emptyArray;
    }

    /**
     * Get list of classes which instances reference this object
     * @param obj referened object
     * @return array of classes which instances reference this object
     * in instance count decreasing order
     */
    public Instance[] referencedByClass(Object obj) { 
        ObjectHeader hdr = (ObjectHeader)objects.get(obj);
        HashMap map = new HashMap();
        if (hdr != null) { 
            Iterator iterator = hdr.referencedBy.iterator();
            while (iterator.hasNext()) { 
                Class c = iterator.next().getClass();
                Instance i = (Instance)map.get(c);
                if (i == null) { 
                    i = new Instance(c);
                }
                i.count += 1;
            }
        }
        Instance[] arr = (Instance[])map.values().toArray(new Instance[map.size()]);
        Arrays.sort(arr);
        return arr;
    }

    /**
     * Find links between objects matching search criteria
     * @param src condition for referencing object
     * @param dst condition for referened object
     * @return array of Link
     */
    public Link[] links(Selector src, Selector dst) { 
        ArrayList result = new ArrayList();
        for (ObjectHeader hdr = blackList; hdr != null; hdr = hdr.next) { 
            Object target = hdr.target;
            if (dst.select(target)) { 
                Iterator iterator = hdr.referencedBy.iterator();
                while (iterator.hasNext()) { 
                    result.add(new Link(iterator.next(), target));
                }
            }
        }
        return (Link[])result.toArray(new Link[result.size()]);                
    }

    /**
     * Find relations between classes matching search criteria
     * @param src condition for referencing object
     * @param dst condition for referened object
     * @return array of Relation in links count decreasing order
     */
    public Relation[] relations(Selector src, Selector dst) { 
        HashMap map = new HashMap();
        for (ObjectHeader hdr = blackList; hdr != null; hdr = hdr.next) { 
            Object target = hdr.target;
            Class targetClass = target.getClass();
            if (dst.select(target)) { 
                Iterator iterator = hdr.referencedBy.iterator();
                while (iterator.hasNext()) { 
                    Object obj = iterator.next();
                    if (src.select(obj)) { 
                        Relation r1 = new Relation(obj.getClass(), targetClass);
                        Relation r2 = (Relation)map.get(r1);
                        if (r2 == null) { 
                            r2 = r1;
                            map.put(r1, r1);
                        }
                        r2.count += 1;
                    }
                }
            }
        }
        Relation[] arr = (Relation[])map.values().toArray(new Relation[map.size()]);
        Arrays.sort(arr);
        return arr;                
    }
                                                        
    /**
     * Get information about ob oject instances accessible from specified roots.
     * @param selector object selector
     * @return array of Instance in count descending order
     */
    public Instance[] instances(Selector selector) { 
        HashMap map = new HashMap();
        for (ObjectHeader hdr = blackList; hdr != null; hdr = hdr.next) { 
            Object target = hdr.target;
            if (selector.select(target)) { 
                Class cls = target.getClass();
                Instance inst = (Instance)map.get(cls);
                if (inst == null) { 
                    inst = new Instance(cls);
                    map.put(cls, inst);
                }
                inst.count += 1;
                if (cls.isArray()) { 
                    inst.totalLength += Array.getLength(target);
                }
            }
        }
        Instance[] arr = (Instance[])map.values().toArray(new Instance[map.size()]);
        Arrays.sort(arr);
        return arr;
    }
        
    private final void add(Object obj, Object from) { 
        if (obj != null) { 
            ObjectHeader hdr = (ObjectHeader)objects.get(obj);
            if (hdr == null) { 
                hdr = new ObjectHeader(obj);
                hdr.next = greyList;
                greyList = hdr;
                //System.out.println("Put " + obj.getClass());
                objects.put(obj, hdr);
            }
            hdr.referencedBy.add(from);
        }
    }
        
    private final void addStaticFields(Class cls) { 
        try { 
            do { 
                Field[] fields = getFields(cls);
                for (int i = 0; i < fields.length; i++) { 
                    Field f = fields[i];
                    if ((f.getModifiers() & Modifier.STATIC) != 0) { 
                        Class type = f.getType();
                        if (!type.isPrimitive()) { 
                            add(f.get(null), cls);
                        }
                    }
                }
            } while ((cls = cls.getSuperclass()) != null);
        } catch (Exception x) { 
            x.printStackTrace();
        }
    }

    private final void mark() {
        while (greyList != null) { 
            ObjectHeader hdr = greyList;
            greyList = hdr.next;
            hdr.next = blackList;
            blackList = hdr;            
            addReferencedObjects(hdr);
        }
    }


    private void addReferencedObjects(ObjectHeader hdr) { 
        Object obj = hdr.target;
        if (obj instanceof Object[]) { 
            Object[] arr = (Object[])obj;
            for (int i = 0; i < arr.length; i++) { 
                add(arr[i], obj);
            }
        } else if (!(obj instanceof Class)) { 
            Class cls = obj.getClass();
            try { 
                do { 
                    Field[] fields = getFields(cls);
                    for (int i = 0; i < fields.length; i++) { 
                        Field f = fields[i];
                        Class type = f.getType();
                        if (!type.isPrimitive()) { 
                            if ((f.getModifiers() & Modifier.STATIC) != 0) { 
                                add(f.get(obj), cls);
                            } else { 
                                add(f.get(obj), obj);
                            }
                        }
                    }
                } while ((cls = cls.getSuperclass()) != null);
            } catch (Exception x) { 
                x.printStackTrace();
            }
        }
    }        

    private Field[] getFields(Class cls) throws Exception
    { 
        Field[] fields = (Field[])classDescMap.get(cls);
        if (fields == null) {
            fields = cls.getDeclaredFields();
            classDescMap.put(cls, fields);
            for (int i = 0; i < fields.length; i++) { 
                fields[i].setAccessible(true);
            }
        }
        return fields;
    }

    static class ObjectHeader { 
        ObjectHeader next;
        boolean      marked;
        Object       target;
        ArrayList    referencedBy;

        ObjectHeader(Object target) { 
            this.target = target;
            referencedBy = new ArrayList();
        }
    }

    private ArrayList rootObjects = new ArrayList();
    private ArrayList rootClasses = new ArrayList();
    private ObjectHeader greyList;
    private ObjectHeader blackList;
    private IdentityHashMap objects = new IdentityHashMap();
    
    private static HashMap classDescMap = new HashMap();
    private static Object[] emptyArray = new Object[0];
}
     